use flare_shader::{Color, IfsConstants, Transform};
use futures_executor::block_on;
use glam::Vec4;
use std::sync::Arc;
use bytemuck::Zeroable;
use wgpu::util::DeviceExt;
use wgpu::{
    Adapter, BindGroupLayout, Device, Features, Instance, Queue, ShaderModule, Surface,
    SurfaceConfiguration,
};
use winit::event::MouseButton;
use winit::{
    application::ApplicationHandler,
    dpi::LogicalSize,
    event::WindowEvent,
    event_loop::{ActiveEventLoop, EventLoop},
    window::{Window, WindowAttributes, WindowId},
};

struct AccumulatePipeline {
    device: Device,
    bind_group_layout: BindGroupLayout,
    pipeline_layout: wgpu::PipelineLayout,
    pipeline: wgpu::ComputePipeline,

    transforms: Option<wgpu::Buffer>,
    accum_image: Option<wgpu::Buffer>,
    bind_group: Option<wgpu::BindGroup>,
}

impl AccumulatePipeline {
    const BIND_GROUP_LAYOUT_DESCRIPTOR: wgpu::BindGroupLayoutDescriptor<'static> =
        wgpu::BindGroupLayoutDescriptor {
            label: Some("accumulate"),
            entries: &[
                // transforms
                wgpu::BindGroupLayoutEntry {
                    binding: 0,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // accum_image
                wgpu::BindGroupLayoutEntry {
                    binding: 1,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: false },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
            ],
        };

    fn create_bind_group(
        device: &Device,
        layout: &wgpu::BindGroupLayout,
        transforms: &wgpu::Buffer,
        accum_image: &wgpu::Buffer,
    ) -> wgpu::BindGroup {
        device.create_bind_group(&wgpu::BindGroupDescriptor {
            label: Some("accumulate"),
            layout,
            entries: &[
                wgpu::BindGroupEntry {
                    binding: 0,
                    resource: transforms.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 1,
                    resource: accum_image.as_entire_binding(),
                },
            ],
        })
    }

    pub fn new(device: &Device, module: &ShaderModule) -> Self {
        let bind_group_layout =
            device.create_bind_group_layout(&Self::BIND_GROUP_LAYOUT_DESCRIPTOR);
        let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("accumulate"),
            bind_group_layouts: &[&bind_group_layout],
            push_constant_ranges: &[wgpu::PushConstantRange {
                stages: wgpu::ShaderStages::COMPUTE,
                range: 0..size_of::<IfsConstants>() as u32,
            }],
        });
        let pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
            label: Some("accumulate"),
            layout: Some(&pipeline_layout),
            module,
            entry_point: Some("main_cs"),
            compilation_options: Default::default(),
            cache: None,
        });

        Self {
            device: device.clone(),
            bind_group_layout,
            pipeline_layout,
            pipeline,
            transforms: None,
            accum_image: None,
            bind_group: None,
        }
    }

    pub fn set_transforms(&mut self, transforms: &[Transform]) {
        // Should be smarter about allocation in the future
        self.transforms = Some(
            self.device
                .create_buffer_init(&wgpu::util::BufferInitDescriptor {
                    label: Some("accumulate/transforms"),
                    contents: bytemuck::cast_slice(transforms),
                    usage: wgpu::BufferUsages::STORAGE,
                }),
        );

        // Setting a new buffer invalidates the existing bindings
        self.bind_group.take();
    }

    pub fn set_accum_image(&mut self, accum_image: &wgpu::Buffer) {
        self.accum_image = Some(accum_image.clone());

        // Setting a new buffer invalidates the existing bindings
        self.bind_group.take();
    }

    fn fetch_bind_group(&mut self) -> &wgpu::BindGroup {
        self.bind_group
            .get_or_insert_with(|| {
                self.device.create_bind_group(&wgpu::BindGroupDescriptor {
                    label: Some("accumulate"),
                    layout: &self.bind_group_layout,
                    entries: &[
                        wgpu::BindGroupEntry {
                            binding: 0,
                            resource: self
                                .transforms
                                .as_ref()
                                .expect("transforms missing")
                                .as_entire_binding(),
                        },
                        wgpu::BindGroupEntry {
                            binding: 1,
                            resource: self
                                .accum_image
                                .as_ref()
                                .expect("accum_image missing")
                                .as_entire_binding(),
                        },
                    ],
                })
            })
    }

    pub fn run(&mut self, encoder: &mut wgpu::CommandEncoder, constants: &IfsConstants) {
        let mut pass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor {
            label: Some("accumulate"),
            timestamp_writes: None,
        });
        pass.set_pipeline(&self.pipeline);
        pass.set_push_constants(0, bytemuck::cast_slice(&[*constants]));
        pass.set_bind_group(0, self.fetch_bind_group(), &[]);
        pass.dispatch_workgroups(1, 1, 1);
    }
}

struct RenderPipeline {
    device: Device,
    bind_group_layout: wgpu::BindGroupLayout,
    pipeline: wgpu::RenderPipeline,

    accum_image: Option<wgpu::Buffer>,
    bind_group: Option<wgpu::BindGroup>,
}

impl RenderPipeline {
    const BIND_GROUP_LAYOUT_DESCRIPTOR: wgpu::BindGroupLayoutDescriptor<'static> =
        wgpu::BindGroupLayoutDescriptor {
            label: Some("render"),
            entries: &[wgpu::BindGroupLayoutEntry {
                binding: 0,
                visibility: wgpu::ShaderStages::FRAGMENT,
                ty: wgpu::BindingType::Buffer {
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                    has_dynamic_offset: false,
                    min_binding_size: None,
                },
                count: None,
            }],
        };

    pub fn new(device: &Device, module: &ShaderModule, format: wgpu::TextureFormat) -> Self {
        let bind_group_layout =
            device.create_bind_group_layout(&Self::BIND_GROUP_LAYOUT_DESCRIPTOR);
        let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("render"),
            bind_group_layouts: &[&bind_group_layout],
            push_constant_ranges: &[wgpu::PushConstantRange {
                stages: wgpu::ShaderStages::FRAGMENT,
                range: 0..size_of::<IfsConstants>() as u32,
            }],
        });
        let pipeline = device.create_render_pipeline(&wgpu::RenderPipelineDescriptor {
            label: Some("render"),
            layout: Some(&pipeline_layout),
            vertex: wgpu::VertexState {
                module,
                entry_point: Some("main_vs"),
                compilation_options: Default::default(),
                buffers: &[],
            },
            primitive: Default::default(),
            depth_stencil: None,
            multisample: Default::default(),
            fragment: Some(wgpu::FragmentState {
                module,
                entry_point: Some("main_fs"),
                compilation_options: Default::default(),
                targets: &[Some(wgpu::ColorTargetState {
                    format,
                    blend: None,
                    write_mask: Default::default(),
                })],
            }),
            multiview: None,
            cache: None,
        });
        Self {
            device: device.clone(),
            bind_group_layout,
            pipeline,
            accum_image: None,
            bind_group: None,
        }
    }

    pub fn set_accum_image(&mut self, accum_image: &wgpu::Buffer) {
        self.accum_image = Some(accum_image.clone());
        self.bind_group.take();
    }

    fn fetch_bind_group(&mut self) -> &wgpu::BindGroup {
        self.bind_group
            .get_or_insert_with(|| {
                self.device.create_bind_group(&wgpu::BindGroupDescriptor {
                    label: Some("render"),
                    layout: &self.bind_group_layout,
                    entries: &[wgpu::BindGroupEntry {
                        binding: 0,
                        resource: self
                            .accum_image
                            .as_ref()
                            .expect("accum_image missing")
                            .as_entire_binding(),
                    }],
                })
            })
    }

    pub fn run(
        &mut self,
        encoder: &mut wgpu::CommandEncoder,
        output_view: &wgpu::TextureView,
        constants: &IfsConstants,
    ) {
        let mut pass = encoder.begin_render_pass(&wgpu::RenderPassDescriptor {
            label: Some("render"),
            color_attachments: &[Some(wgpu::RenderPassColorAttachment {
                view: output_view,
                resolve_target: None,
                ops: Default::default(),
            })],
            depth_stencil_attachment: None,
            timestamp_writes: None,
            occlusion_query_set: None,
        });
        pass.set_pipeline(&self.pipeline);
        pass.set_push_constants(
            wgpu::ShaderStages::FRAGMENT,
            0,
            bytemuck::cast_slice(&[*constants]),
        );
        pass.set_bind_group(0, self.fetch_bind_group(), &[]);
        pass.draw(0..3, 0..1);
    }
}

struct Flare {
    instance: Instance,
    adapter: Adapter,
    device: Device,
    queue: Queue,
    module: ShaderModule,
}

impl Flare {
    pub fn new() -> Self {
        let backends = wgpu::Backends::from_env().unwrap_or_default();
        let instance = Instance::new(&wgpu::InstanceDescriptor {
            backends,
            ..Default::default()
        });

        let adapter = instance.request_adapter(&wgpu::RequestAdapterOptions {
            power_preference: wgpu::PowerPreference::HighPerformance,
            force_fallback_adapter: false,
            compatible_surface: None,
        });
        let adapter = block_on(adapter).expect("Failed to find GPU adapter");

        let required_limits = wgpu::Limits {
            max_push_constant_size: size_of::<IfsConstants>() as u32,
            ..Default::default()
        };

        let device = adapter.request_device(
            &wgpu::DeviceDescriptor {
                label: Some("flare"),
                required_features: Features::TIMESTAMP_QUERY | Features::PUSH_CONSTANTS,
                required_limits,
                memory_hints: Default::default(),
            },
            None,
        );
        let (device, queue) = block_on(device).expect("Failed to find GPU device");
        let module = device
            .create_shader_module(wgpu::include_spirv!(concat!(env!("OUT_DIR"), "/flare.spv")));

        Flare {
            instance,
            adapter,
            device,
            queue,
            module,
        }
    }
}

struct FlareRender<'window> {
    flare: Arc<Flare>,
    surface: Surface<'window>,
    surface_configuration: SurfaceConfiguration,
    accumulate_pipeline: AccumulatePipeline,
    render_pipeline: RenderPipeline,
    ifs_constants: IfsConstants,
    accum_image: wgpu::Buffer,
}

impl FlareRender<'_> {
    fn create_accum_image(device: &Device, width: u32, height: u32) -> wgpu::Buffer {
        let pixels = (width * height) as u64;
        let pixel_size = 4u64 * size_of::<f32>() as u64;
        let size = pixels * pixel_size;
        device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("accum_image"),
            size,
            usage: wgpu::BufferUsages::STORAGE,
            mapped_at_creation: false,
        })
    }

    pub fn new(flare: Arc<Flare>, window: Arc<Window>) -> Self {
        let window_size = window.inner_size();

        let surface = flare
            .instance
            .create_surface(window.clone())
            .expect("Unable to create surface");
        let mut surface_configuration = surface
            .get_default_config(&flare.adapter, window_size.width, window_size.height)
            .expect("Unable to get surface config");
        surface_configuration.present_mode = wgpu::PresentMode::AutoVsync;
        surface.configure(&flare.device, &surface_configuration);

        let mut accumulate_pipeline = AccumulatePipeline::new(&flare.device, &flare.module);
        let mut render_pipeline =
            RenderPipeline::new(&flare.device, &flare.module, surface_configuration.format);

        let ifs_constants = IfsConstants::new(
            window_size.width,
            window_size.height,
            window_size.width,
            window_size.height,
            Vec4::BLACK,
        );

        let accum_image = Self::create_accum_image(&flare.device, window_size.width, window_size.height);

        accumulate_pipeline.set_transforms(&[Transform::zeroed()]);
        accumulate_pipeline.set_accum_image(&accum_image);
        render_pipeline.set_accum_image(&accum_image);

        Self {
            flare,
            surface,
            surface_configuration,
            accumulate_pipeline,
            render_pipeline,
            ifs_constants,
            accum_image,
        }
    }

    pub fn render(&mut self, run_accumulate: bool) {
        let output = self
            .surface
            .get_current_texture()
            .expect("Failed to get current texture");
        let output_view = output.texture.create_view(&Default::default());
        let mut encoder = self
            .flare
            .device
            .create_command_encoder(&Default::default());

        if run_accumulate {
            self.accumulate_pipeline.run(&mut encoder, &self.ifs_constants);
        }
        self.render_pipeline.run(&mut encoder, &output_view, &self.ifs_constants);

        self.flare.queue.submit(Some(encoder.finish()));
        output.present();
    }

    pub fn resize_accumulate(&mut self) {
        let vp_dimensions = self.ifs_constants.viewport_dimensions();
        self.ifs_constants
            .with_accumulate(vp_dimensions.x, vp_dimensions.y);

        let accum_image = Self::create_accum_image(&self.flare.device, vp_dimensions.x, vp_dimensions.y);
        self.accumulate_pipeline.set_accum_image(&accum_image);
        self.render_pipeline.set_accum_image(&accum_image);
        self.accum_image = accum_image;
    }

    pub fn resize_viewport(&mut self, width: u32, height: u32) {
        self.surface_configuration.width = width;
        self.surface_configuration.height = height;
        self.surface
            .configure(&self.flare.device, &self.surface_configuration);
        self.ifs_constants.with_viewport(width, height);
    }
}

struct Application<'window> {
    flare: Arc<Flare>,
    window: Option<Arc<Window>>,
    flare_render: Option<FlareRender<'window>>,
}

impl Application<'_> {
    pub fn new() -> Self {
        Self {
            flare: Arc::new(Flare::new()),
            window: None,
            flare_render: None,
        }
    }
}

impl ApplicationHandler for Application<'_> {
    fn resumed(&mut self, event_loop: &ActiveEventLoop) {
        let attributes = WindowAttributes::default()
            .with_title("Flare")
            .with_inner_size(LogicalSize::new(1024, 768));

        let window = Arc::new(
            event_loop
                .create_window(attributes)
                .expect("Failed to create window"),
        );

        self.window = Some(window.clone());

        let mut flare_render = FlareRender::new(self.flare.clone(), window);
        flare_render.render(true);

        self.flare_render = Some(flare_render);
    }

    fn window_event(
        &mut self,
        event_loop: &ActiveEventLoop,
        _window_id: WindowId,
        event: WindowEvent,
    ) {
        match event {
            WindowEvent::Resized(size) => {
                let flare_render = self.flare_render.as_mut().unwrap();
                flare_render.resize_viewport(size.width, size.height);
                flare_render.render(false);
            }
            WindowEvent::MouseInput {
                button: MouseButton::Left,
                ..
            } => {
                let flare_render = self.flare_render.as_mut().unwrap();
                flare_render.resize_accumulate();
                flare_render.render(true);
            }
            WindowEvent::CloseRequested => event_loop.exit(),
            _ => (),
        }
    }
}

pub fn main() -> anyhow::Result<()> {
    env_logger::init();
    let _module = wgpu::include_spirv!(concat!(env!("OUT_DIR"), "/flare.spv"));

    let event_loop = EventLoop::new()?;
    let mut application = Application::new();
    event_loop.run_app(&mut application)?;

    Ok(())
}
